﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.Sockets;
using System.Threading;

namespace SocketAsyncOpWithTimeOut
{
    public static class SocketExtension
    {

        public static int EndReceive2(IAsyncResult ar)
        {
            var result = ar as AsyncResultWithTimeout;
            result.AsyncWaitHandle.WaitOne();

            return result.ReceivedCount;
        }

        public static AsyncResultWithTimeout BeginReceive2
        (
            this Socket socket,
            int timeout,
            byte[] buffer,
            int offset,
            int size,
            SocketFlags flags,
            AsyncCallback callback,
            object state
        )
        {
            var result = new AsyncResultWithTimeout(callback, state);

            var stateInfo = new StateInfo(result, socket);

            socket.BeginReceive(buffer, offset, size, flags, ReceiveCb, state);

            var registeredWaitHandle =
                ThreadPool.RegisterWaitForSingleObject(
                    result.AsyncWaitHandle,
                    WatchTimeOut,
                    stateInfo, // 作为state传递给WatchTimeOut
                    timeout,
                    true);

            // stateInfo中保存RegisteredWaitHandle，以方便在WatchTimeOut
            // 中unregister.
            stateInfo.RegisteredWaitHandle = registeredWaitHandle;

            return result;
        }

        private static void WatchTimeOut(object state, bool timeout)
        {
            var stateInfo = state as StateInfo;
            // 设置的timeout前，操作未完成，则设置为操作Timeout
            if (timeout)
            {
                stateInfo.AsycResult.SetTimeout();
            }

            // 取消之前注册的WaitOrTimerCallback
            stateInfo.RegisteredWaitHandle.Unregister(
                stateInfo.AsycResult.AsyncWaitHandle);
        }

        private static void ReceiveCb(IAsyncResult result)
        {
            var state = result.AsyncState as StateInfo;
            var asyncResultWithTimeOut = state.AsycResult;
            var count = state.Socket.EndReceive(result);
            state.AsycResult.SetResult(count);
        }
    }


    public class StateInfo
    {
        public StateInfo(AsyncResultWithTimeout result, Socket socket)
        {
            this.AsycResult = result;
            this.Socket = socket;
        }

        public Socket Socket { get; private set; }
        public AsyncResultWithTimeout AsycResult { get; private set; }
        public RegisteredWaitHandle RegisteredWaitHandle { get; set; }
    }

    public class AsyncResultWithTimeout : IAsyncResult
    {
        private ManualResetEvent m_waitHandle = new ManualResetEvent(false);
        public AsyncResultWithTimeout(AsyncCallback cb, object state)
        {
            this.AsyncState = state;
            this.Callback = cb;
        }

        #region IAsyncResult

        public object AsyncState { get; private set; }
        public WaitHandle AsyncWaitHandle { get { return m_waitHandle; } }
        public bool CompletedSynchronously { get { return false; } }
        public bool IsCompleted { get; private set; }

        #endregion

        public AsyncCallback Callback { get; private set; }
        public int ReceivedCount { get; private set; }
        public bool TimedOut { get; private set; }
        public void SetResult(int count)
        {
            this.IsCompleted = true;
            this.ReceivedCount = count;
            this.m_waitHandle.Set();

            if (Callback != null) Callback(this);
        }

        public void SetTimeout()
        {
            this.TimedOut = true;
            this.IsCompleted = true;
            this.m_waitHandle.Set();
        }
    }
}